@@ -1,5 +1,5 @@
 /*
- * $PostgreSQL: pgsql/contrib/xml2/xpath.c,v 1.23.2.4 2010/03/03 19:10:29 tgl Exp $
+ * contrib/xml2/xpath.c
  *
  * Parser interface for DOM-based parser (libxml) rather than
  * stream-based SAX-type parser
@@ -38,7 +38,7 @@ Datum		xpath_table(PG_FUNCTION_ARGS);
 
 /* exported for use by xslt_proc.c */
 
-void		pgxml_parser_init(void);
+PgXmlErrorContext *pgxml_parser_init(PgXmlStrictness strictness);
 
 /* workspace for pgxml_xpath() */
 
@@ -61,29 +61,45 @@ static text *pgxml_result_to_text(xmlXPathObjectPtr res, xmlChar *toptag,
 static xmlChar *pgxml_texttoxmlchar(text *textstring);
 
 static xmlXPathObjectPtr pgxml_xpath(text *document, xmlChar *xpath,
-									 xpath_workspace *workspace);
+			xpath_workspace *workspace);
 
 static void cleanup_workspace(xpath_workspace *workspace);
 
 
 /*
  * Initialize for xml parsing.
+ *
+ * As with the underlying pg_xml_init function, calls to this MUST be followed
+ * by a PG_TRY block that guarantees that pg_xml_done is called.
  */
-void
-pgxml_parser_init(void)
+PgXmlErrorContext *
+pgxml_parser_init(PgXmlStrictness strictness)
 {
+	PgXmlErrorContext *xmlerrcxt;
+
 	/* Set up error handling (we share the core's error handler) */
-	pg_xml_init();
+	xmlerrcxt = pg_xml_init(strictness);
+
+	/* Note: we're assuming an elog cannot be thrown by the following calls */
 
 	/* Initialize libxml */
 	xmlInitParser();
 
 	xmlSubstituteEntitiesDefault(1);
 	xmlLoadExtDtdDefaultValue = 1;
+
+	return xmlerrcxt;
 }
 
 
-/* Returns true if document is well-formed */
+/*
+ * Returns true if document is well-formed
+ *
+ * Note: this has been superseded by a core function.  We still have to
+ * have it in the contrib module so that existing SQL-level references
+ * to the function won't fail; but in normal usage with up-to-date SQL
+ * definitions for the contrib module, this won't be called.
+ */
 
 PG_FUNCTION_INFO_V1(xml_is_well_formed);
 
@@ -91,16 +107,33 @@ Datum
 xml_is_well_formed(PG_FUNCTION_ARGS)
 {
 	text	   *t = PG_GETARG_TEXT_P(0);		/* document buffer */
+	bool		result = false;
 	int32		docsize = VARSIZE(t) - VARHDRSZ;
 	xmlDocPtr	doctree;
+	PgXmlErrorContext *xmlerrcxt;
 
-	pgxml_parser_init();
+	xmlerrcxt = pgxml_parser_init(PG_XML_STRICTNESS_LEGACY);
 
-	doctree = xmlParseMemory((char *) VARDATA(t), docsize);
-	if (doctree == NULL)
-		PG_RETURN_BOOL(false);	/* i.e. not well-formed */
-	xmlFreeDoc(doctree);
-	PG_RETURN_BOOL(true);
+	PG_TRY();
+	{
+		doctree = xmlParseMemory((char *) VARDATA(t), docsize);
+
+		result = (doctree != NULL);
+
+		if (doctree != NULL)
+			xmlFreeDoc(doctree);
+	}
+	PG_CATCH();
+	{
+		pg_xml_done(xmlerrcxt, true);
+
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
+
+	pg_xml_done(xmlerrcxt, false);
+
+	PG_RETURN_BOOL(result);
 }
 
 
@@ -227,7 +260,7 @@ Datum
 xpath_nodeset(PG_FUNCTION_ARGS)
 {
 	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+	text	   *xpathsupp = PG_GETARG_TEXT_P(1);		/* XPath expression */
 	xmlChar    *toptag = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(2));
 	xmlChar    *septag = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(3));
 	xmlChar    *xpath;
@@ -260,7 +293,7 @@ Datum
 xpath_list(PG_FUNCTION_ARGS)
 {
 	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+	text	   *xpathsupp = PG_GETARG_TEXT_P(1);		/* XPath expression */
 	xmlChar    *plainsep = pgxml_texttoxmlchar(PG_GETARG_TEXT_P(2));
 	xmlChar    *xpath;
 	text	   *xpres;
@@ -289,7 +322,7 @@ Datum
 xpath_string(PG_FUNCTION_ARGS)
 {
 	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+	text	   *xpathsupp = PG_GETARG_TEXT_P(1);		/* XPath expression */
 	xmlChar    *xpath;
 	int32		pathsize;
 	text	   *xpres;
@@ -330,7 +363,7 @@ Datum
 xpath_number(PG_FUNCTION_ARGS)
 {
 	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+	text	   *xpathsupp = PG_GETARG_TEXT_P(1);		/* XPath expression */
 	xmlChar    *xpath;
 	float4		fRes;
 	xmlXPathObjectPtr res;
@@ -362,7 +395,7 @@ Datum
 xpath_bool(PG_FUNCTION_ARGS)
 {
 	text	   *document = PG_GETARG_TEXT_P(0);
-	text	   *xpathsupp = PG_GETARG_TEXT_P(1);	/* XPath expression */
+	text	   *xpathsupp = PG_GETARG_TEXT_P(1);		/* XPath expression */
 	xmlChar    *xpath;
 	int			bRes;
 	xmlXPathObjectPtr res;
@@ -392,41 +425,52 @@ static xmlXPathObjectPtr
 pgxml_xpath(text *document, xmlChar *xpath, xpath_workspace *workspace)
 {
 	int32		docsize = VARSIZE(document) - VARHDRSZ;
-	xmlXPathObjectPtr res;
+	PgXmlErrorContext *xmlerrcxt;
 	xmlXPathCompExprPtr comppath;
 
 	workspace->doctree = NULL;
 	workspace->ctxt = NULL;
 	workspace->res = NULL;
 
-	pgxml_parser_init();
+	xmlerrcxt = pgxml_parser_init(PG_XML_STRICTNESS_LEGACY);
 
-	workspace->doctree = xmlParseMemory((char *) VARDATA(document), docsize);
-	if (workspace->doctree == NULL)
-		return NULL;			/* not well-formed */
+	PG_TRY();
+	{
+		workspace->doctree = xmlParseMemory((char *) VARDATA(document),
+											docsize);
+		if (workspace->doctree != NULL)
+		{
+			workspace->ctxt = xmlXPathNewContext(workspace->doctree);
+			workspace->ctxt->node = xmlDocGetRootElement(workspace->doctree);
 
-	workspace->ctxt = xmlXPathNewContext(workspace->doctree);
-	workspace->ctxt->node = xmlDocGetRootElement(workspace->doctree);
+			/* compile the path */
+			comppath = xmlXPathCompile(xpath);
+			if (comppath == NULL)
+				xml_ereport(xmlerrcxt, ERROR, ERRCODE_EXTERNAL_ROUTINE_EXCEPTION,
+							"XPath Syntax Error");
 
-	/* compile the path */
-	comppath = xmlXPathCompile(xpath);
-	if (comppath == NULL)
+			/* Now evaluate the path expression. */
+			workspace->res = xmlXPathCompiledEval(comppath, workspace->ctxt);
+
+			xmlXPathFreeCompExpr(comppath);
+		}
+	}
+	PG_CATCH();
 	{
 		cleanup_workspace(workspace);
-		xml_ereport(ERROR, ERRCODE_EXTERNAL_ROUTINE_EXCEPTION,
-					"XPath Syntax Error");
+
+		pg_xml_done(xmlerrcxt, true);
+
+		PG_RE_THROW();
 	}
+	PG_END_TRY();
 
-	/* Now evaluate the path expression. */
-	res = xmlXPathCompiledEval(comppath, workspace->ctxt);
-	workspace->res = res;
-
-	xmlXPathFreeCompExpr(comppath);
-
-	if (res == NULL)
+	if (workspace->res == NULL)
 		cleanup_workspace(workspace);
 
-	return res;
+	pg_xml_done(xmlerrcxt, false);
+
+	return workspace->res;
 }
 
 /* Clean up after processing the result of pgxml_xpath() */
@@ -527,6 +571,8 @@ xpath_table(PG_FUNCTION_ARGS)
 								 * document */
 	bool		had_values;		/* To determine end of nodeset results */
 	StringInfoData query_buf;
+	PgXmlErrorContext *xmlerrcxt;
+	volatile xmlDocPtr doctree = NULL;
 
 	/* We only have a valid tuple description in table function mode */
 	if (rsinfo == NULL || !IsA(rsinfo, ReturnSetInfo))
@@ -649,133 +695,149 @@ xpath_table(PG_FUNCTION_ARGS)
 	}
 
 	/*
-	 * Setup the parser.  This should happen after we are done evaluating
-	 * the query, in case it calls functions that set up libxml differently.
+	 * Setup the parser.  This should happen after we are done evaluating the
+	 * query, in case it calls functions that set up libxml differently.
 	 */
-	pgxml_parser_init();
+	xmlerrcxt = pgxml_parser_init(PG_XML_STRICTNESS_LEGACY);
 
-	/* For each row i.e. document returned from SPI */
-	for (i = 0; i < proc; i++)
+	PG_TRY();
 	{
-		char	   *pkey;
-		char	   *xmldoc;
-		xmlDocPtr	doctree;
-		xmlXPathContextPtr ctxt;
-		xmlXPathObjectPtr res;
-		xmlChar    *resstr;
-		xmlXPathCompExprPtr comppath;
+		/* For each row i.e. document returned from SPI */
+		for (i = 0; i < proc; i++)
+		{
+			char	   *pkey;
+			char	   *xmldoc;
+			xmlXPathContextPtr ctxt;
+			xmlXPathObjectPtr res;
+			xmlChar    *resstr;
+			xmlXPathCompExprPtr comppath;
 
-		/* Extract the row data as C Strings */
-		spi_tuple = tuptable->vals[i];
-		pkey = SPI_getvalue(spi_tuple, spi_tupdesc, 1);
-		xmldoc = SPI_getvalue(spi_tuple, spi_tupdesc, 2);
+			/* Extract the row data as C Strings */
+			spi_tuple = tuptable->vals[i];
+			pkey = SPI_getvalue(spi_tuple, spi_tupdesc, 1);
+			xmldoc = SPI_getvalue(spi_tuple, spi_tupdesc, 2);
 
-		/*
-		 * Clear the values array, so that not-well-formed documents return
-		 * NULL in all columns.  Note that this also means that spare columns
-		 * will be NULL.
-		 */
-		for (j = 0; j < ret_tupdesc->natts; j++)
-			values[j] = NULL;
+			/*
+			 * Clear the values array, so that not-well-formed documents
+			 * return NULL in all columns.  Note that this also means that
+			 * spare columns will be NULL.
+			 */
+			for (j = 0; j < ret_tupdesc->natts; j++)
+				values[j] = NULL;
 
-		/* Insert primary key */
-		values[0] = pkey;
+			/* Insert primary key */
+			values[0] = pkey;
 
-		/* Parse the document */
-		if (xmldoc)
-			doctree = xmlParseMemory(xmldoc, strlen(xmldoc));
-		else	/* treat NULL as not well-formed */
+			/* Parse the document */
+			if (xmldoc)
+				doctree = xmlParseMemory(xmldoc, strlen(xmldoc));
+			else	/* treat NULL as not well-formed */
+				doctree = NULL;
+
+			if (doctree == NULL)
+			{
+				/* not well-formed, so output all-NULL tuple */
+				ret_tuple = BuildTupleFromCStrings(attinmeta, values);
+				tuplestore_puttuple(tupstore, ret_tuple);
+				heap_freetuple(ret_tuple);
+			}
+			else
+			{
+				/* New loop here - we have to deal with nodeset results */
+				rownr = 0;
+
+				do
+				{
+					/* Now evaluate the set of xpaths. */
+					had_values = false;
+					for (j = 0; j < numpaths; j++)
+					{
+						ctxt = xmlXPathNewContext(doctree);
+						ctxt->node = xmlDocGetRootElement(doctree);
+
+						/* compile the path */
+						comppath = xmlXPathCompile(xpaths[j]);
+						if (comppath == NULL)
+							xml_ereport(xmlerrcxt, ERROR,
+										ERRCODE_EXTERNAL_ROUTINE_EXCEPTION,
+										"XPath Syntax Error");
+
+						/* Now evaluate the path expression. */
+						res = xmlXPathCompiledEval(comppath, ctxt);
+						xmlXPathFreeCompExpr(comppath);
+
+						if (res != NULL)
+						{
+							switch (res->type)
+							{
+								case XPATH_NODESET:
+									/* We see if this nodeset has enough nodes */
+									if (res->nodesetval != NULL &&
+										rownr < res->nodesetval->nodeNr)
+									{
+										resstr = xmlXPathCastNodeToString(res->nodesetval->nodeTab[rownr]);
+										had_values = true;
+									}
+									else
+										resstr = NULL;
+
+									break;
+
+								case XPATH_STRING:
+									resstr = xmlStrdup(res->stringval);
+									break;
+
+								default:
+									elog(NOTICE, "unsupported XQuery result: %d", res->type);
+									resstr = xmlStrdup((const xmlChar *) "<unsupported/>");
+							}
+
+							/*
+							 * Insert this into the appropriate column in the
+							 * result tuple.
+							 */
+							values[j + 1] = (char *) resstr;
+						}
+						xmlXPathFreeContext(ctxt);
+					}
+
+					/* Now add the tuple to the output, if there is one. */
+					if (had_values)
+					{
+						ret_tuple = BuildTupleFromCStrings(attinmeta, values);
+						tuplestore_puttuple(tupstore, ret_tuple);
+						heap_freetuple(ret_tuple);
+					}
+
+					rownr++;
+				} while (had_values);
+			}
+
+			if (doctree != NULL)
+				xmlFreeDoc(doctree);
 			doctree = NULL;
 
-		if (doctree == NULL)
-		{
-			/* not well-formed, so output all-NULL tuple */
-			ret_tuple = BuildTupleFromCStrings(attinmeta, values);
-			tuplestore_puttuple(tupstore, ret_tuple);
-			heap_freetuple(ret_tuple);
+			if (pkey)
+				pfree(pkey);
+			if (xmldoc)
+				pfree(xmldoc);
 		}
-		else
-		{
-			/* New loop here - we have to deal with nodeset results */
-			rownr = 0;
+	}
+	PG_CATCH();
+	{
+		if (doctree != NULL)
+			xmlFreeDoc(doctree);
 
-			do
-			{
-				/* Now evaluate the set of xpaths. */
-				had_values = false;
-				for (j = 0; j < numpaths; j++)
-				{
-					ctxt = xmlXPathNewContext(doctree);
-					ctxt->node = xmlDocGetRootElement(doctree);
+		pg_xml_done(xmlerrcxt, true);
 
-					/* compile the path */
-					comppath = xmlXPathCompile(xpaths[j]);
-					if (comppath == NULL)
-					{
-						xmlFreeDoc(doctree);
-						xml_ereport(ERROR, ERRCODE_EXTERNAL_ROUTINE_EXCEPTION,
-									"XPath Syntax Error");
-					}
-
-					/* Now evaluate the path expression. */
-					res = xmlXPathCompiledEval(comppath, ctxt);
-					xmlXPathFreeCompExpr(comppath);
-
-					if (res != NULL)
-					{
-						switch (res->type)
-						{
-							case XPATH_NODESET:
-								/* We see if this nodeset has enough nodes */
-								if (res->nodesetval != NULL &&
-									rownr < res->nodesetval->nodeNr)
-								{
-									resstr =
-										xmlXPathCastNodeToString(res->nodesetval->nodeTab[rownr]);
-									had_values = true;
-								}
-								else
-									resstr = NULL;
-
-								break;
-
-							case XPATH_STRING:
-								resstr = xmlStrdup(res->stringval);
-								break;
-
-							default:
-								elog(NOTICE, "unsupported XQuery result: %d", res->type);
-								resstr = xmlStrdup((const xmlChar *) "<unsupported/>");
-						}
-
-						/*
-						 * Insert this into the appropriate column in the
-						 * result tuple.
-						 */
-						values[j + 1] = (char *) resstr;
-					}
-					xmlXPathFreeContext(ctxt);
-				}
-
-				/* Now add the tuple to the output, if there is one. */
-				if (had_values)
-				{
-					ret_tuple = BuildTupleFromCStrings(attinmeta, values);
-					tuplestore_puttuple(tupstore, ret_tuple);
-					heap_freetuple(ret_tuple);
-				}
-
-				rownr++;
-			} while (had_values);
-		}
+		PG_RE_THROW();
+	}
+	PG_END_TRY();
 
+	if (doctree != NULL)
 		xmlFreeDoc(doctree);
 
-		if (pkey)
-			pfree(pkey);
-		if (xmldoc)
-			pfree(xmldoc);
-	}
+	pg_xml_done(xmlerrcxt, false);
 
 	tuplestore_donestoring(tupstore);
 
